import re
from functools import wraps, partial
from collections import defaultdict

from oauth2client.client import OAuth2Credentials, OAuth2WebServerFlow, FlowExchangeError
from twisted.internet import reactor, task
from twisted.internet.defer import Deferred
from urllib2 import HTTPError

from helga import log, settings
from helga.db import db
from helga.plugins import command, match, random_ack, ResponseNotReady


USE_OAUTH = getattr(settings, 'VERSIONONE_OAUTH_ENABLED', False)
if USE_OAUTH:
    from helga_versionone.v1_wrapper import HelgaV1Meta as V1Meta  # pragma: no cover
else:
    from v1pysdk import V1Meta

logger = log.getLogger(__name__)
VERSIONONE_PATTERNS = set(['B', 'D', 'TK', 'AT', 'FG', 'I', 'R', 'E'])
# Non-workitems, need different endpoints
SPECIAL_PATTERNS = {
    'I': 'Issue',
    'R': 'Request',
}


class NotFound(Exception):
    pass


class QuitNow(Exception):
    pass


def bad_auth(v1, client, channel, nick, failure):
    failure.trap(HTTPError)
    logger.debug('bag_auth failure="{0}"'.format(failure))
    client.msg(channel, u'{0}, You probably need to reset your token, try "!v1 token"'.format(nick))


def bad_args(v1, client, channel, nick, failure):
    failure.trap(TypeError, AttributeError, HTTPError)
    logger.debug('bag_args failure="{0}"'.format(failure))

    if v1 is None:
        client.msg(channel, u'{0}, you might want to try "!v1 oauth" or "!v1 token"'.format(nick))
    else:
        logger.warning('Check docs because', exc_info=True)
        client.msg(channel, u'Umm... {0}, you might want to check the docs for that'.format(nick))


def quit_now(client, channel, nick, failure):
    failure.trap(QuitNow)
    client.msg(channel, failure.value.message.format(channel=channel, nick=nick))


class deferred_response(object):
    def __init__(self, target):
        self.target = target

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(v1, client, channel, nick, *args):
            d = task.deferLater(
                reactor, 0, fn, v1, client, channel, nick, *args
            ).addCallback(
                partial(client.msg, locals()[self.target])
            ).addErrback(
                partial(bad_auth, v1, client, channel, nick)
            ).addErrback(
                partial(bad_args, v1, client, channel, nick)
            ).addErrback(
                partial(quit_now, client, channel, nick)
            )
            logger.debug('Delaying exection of {0}'.format(fn.__name__))
            return d
        return wrapper

deferred_to_channel = deferred_response('channel')
deferred_to_nick = deferred_response('nick')


def commit_changes(v1, *args):
    """Respect the READONLY setting, return ack, or no perms message
       args is a list of 3-tuples, that will be passed to setattr iff we will write to V1

       Because of the V1Meta caching, strange things happen if you make changes and don't actually commit
       So only make the changes when commiting.
    """
    if getattr(settings, 'VERSIONONE_READONLY', True):
        return 'I would, but I\'m not allowed to write :('

    for call in args:
        setattr(*call)

    v1.commit()
    return random_ack()


def create_object(Klass, **kwargs):
    """Respect the READONLY setting return the object or raise QuitNow
       kwargs are passed to Klass.create()

    """
    if getattr(settings, 'VERSIONONE_READONLY', True):
        raise QuitNow('I\'m sorry {nick}, write access is disabled')

    return Klass.create(**kwargs)


def _get_things(Klass, workitem, *args):
    if not args:
        args = ['Name', 'Status.Name', 'Status.Order']
    return Klass.where(Parent=workitem.idref).select(*args)


def get_workitem(v1, number, *args):
    """Get a workitem or die trying.
       args are passed to select to pre-populate fields
    """
    try:
        return v1.Workitem.where(Number=number).select(*args).first()
    except IndexError:
        raise QuitNow('I\'m sorry {{nick}}, item "{0}" not found'.format(number))


def get_user(v1, nick):
    o_nick = nick
    try:
        nick = db.v1_user_map.find_one({'irc_nick': nick})['v1_nick']
    except (TypeError, KeyError):
        # Lookup failed, no worries
        pass

    try:
        return v1.Member.filter(
            "Name='{0}'|Nickname='{0}'|Username='{0}'".format(nick)
        ).select(
            'Name', 'Nickname'
        ).first()
    except IndexError:
        raise QuitNow(
            'I\'m sorry {{nick}}, couldn\'t find {0} in VersionOne as {1}. '
            'Check !v1 alias'.format(o_nick, nick)
        )


def get_creds(nick):
    """returns simple token as string, OAuth2Credentials instance, or None
    """

    auth_info = db.v1_oauth.find_one({'irc_nick': nick}) or None
    # Try to trim nick for usual things
    if auth_info is None:
        if '|' in nick:
            nick = nick.split('|', 1)[0].strip()
        elif '_' in nick:
            nick = nick.split('_', 1)[0].strip()
        auth_info = db.v1_oauth.find_one({'irc_nick': nick}) or {}

    token = auth_info.get('api_token')
    if token:
        return token

    if USE_OAUTH:
        try:
            return OAuth2Credentials(
                auth_info['access_token'],
                settings.VERSIONONE_OAUTH_CLIENT_ID,
                settings.VERSIONONE_OAUTH_CLIENT_SECRET,
                auth_info['refresh_token'],
                auth_info['token_expiry'],
                settings.VERSIONONE_URL + '/oauth.v1/token',
                'helga (chatbot)',
            )
        except Exception:
            # TODO - what can get raised here
            logger.warning('Problem getting OAuth creds for {0}'.format(nick), exc_info=True)
            raise QuitNow('Sorry {{nick}}, couldn\'t get OAuth creds for you, try "!v1 oauth"')

    return None


@deferred_to_channel
def alias_command(v1, client, channel, nick, *args):
    # Populate subcmd, and target to continue
    target = nick
    try:
        subcmd = args[0]
        args = args[1:]
    except IndexError:
        # 0 args lookup nick
        subcmd = 'lookup'
    else:
        # At least 1 arg
        if args:
            target = ' '.join(args)
        else:
            # Exactly 1 arg - look it up if not a command
            if subcmd not in ['lookup', 'remove']:
                target = subcmd
                subcmd = 'lookup'

    if subcmd == 'lookup':
        lookup = {'irc_nick': target}
        try:
            v1_nick = db.v1_user_map.find_one(lookup)['v1_nick']
        except (TypeError, KeyError):
            v1_nick = target
        return '{0} is known as {1} in V1'.format(target, v1_nick)

    elif subcmd == 'set':
        lookup = {'irc_nick': nick}
        alias = db.v1_user_map.find_one(lookup) or lookup
        alias['v1_nick'] = target
        db.v1_user_map.save(alias)

    elif subcmd == 'remove':
        if target != nick:
            return 'That\'s not nice {0}. You can\'t remove {1}'.format(nick, target)
        lookup = {'irc_nick': nick}
        db.v1_user_map.find_and_modify(lookup, remove=True)
    else:
        return 'No {0}, you can\'t {1}!'.format(nick, subcmd)

    return random_ack()


@deferred_to_nick
def token_command(v1, client, channel, nick, reply_code=None):
    if reply_code:
        q = {'irc_nick': nick}
        auth_info = db.v1_oauth.find_one(q) or q

        if reply_code == 'forget':
            # remove the token
            try:
                del(auth_info['api_token'])
            except KeyError:
                return 'Token was already gone'
        else:
            # update the token
            auth_info['api_token'] = reply_code
        db.v1_oauth.save(auth_info)
        return random_ack()

    # No reply_code - show step1 instructions
    return (
        'In V1 go to your Applications and generate a Personal Access Token'
        'then do "!v1 token <code>" with the generated code'
    )


@deferred_to_nick
def oauth_command(v1, client, channel, nick, reply_code=None):
    if not USE_OAUTH:
        return 'Oauth is not enabled'

    client = OAuth2WebServerFlow(
        settings.VERSIONONE_OAUTH_CLIENT_ID,
        settings.VERSIONONE_OAUTH_CLIENT_SECRET,
        'apiv1',  # Scope for XML api
        redirect_uri='urn:ietf:wg:oauth:2.0:oob',
        auth_uri=settings.VERSIONONE_URL + '/oauth.v1/auth',
        token_uri=settings.VERSIONONE_URL + '/oauth.v1/token',
    )

    if reply_code:
        q = {'irc_nick': nick}
        auth_info = db.v1_oauth.find_one(q) or q

        if reply_code == 'forget':
            for key in ['access_token', 'refresh_token', 'token_expiry']:
                try:
                    del(auth_info[key])
                except KeyError:
                    pass
        else:
            try:
                creds = client.step2_exchange(reply_code)
            except FlowExchangeError as e:
                return 'Sorry {0} "{1}" happened. Try "!v1 oauth" again from the start'.format(nick, e)

            # Creds Ok, save the info
            auth_info['access_token'] = creds.access_token
            auth_info['refresh_token'] = creds.refresh_token
            auth_info['token_expiry'] = creds.token_expiry

        db.v1_oauth.save(auth_info)
        return random_ack()

    # No reply_code - show step1 link
    return 'Visit {0} then do "!v1 oauth <code>" with the generated code'.format(
        client.step1_get_authorize_url())


def get_v1(nick):
    """Get the v1 connection"""

    try:
        credentials = get_creds(nick)

        # Access Token is prefered, remove token to use OAUTH
        if isinstance(credentials, basestring):
            logger.debug('Using V1 token auth')
            v1 = V1Meta(
                instance_url=settings.VERSIONONE_URL,
                password=credentials,
                use_password_as_token=True,
            )

        # Use Oauth if provided
        elif credentials:
            logger.debug('Trying V1 oauth')
            v1 = V1Meta(
                instance_url=settings.VERSIONONE_URL,
                credentials=credentials,
            )

        # System user if no creds
        else:
            # Too much overhead, don't really need to re-init every call
            logger.debug('Trying V1 service user auth')
            v1 = V1Meta(
                instance_url=settings.VERSIONONE_URL,
                username=settings.VERSIONONE_AUTH[0],
                password=settings.VERSIONONE_AUTH[1],
            )

    except AttributeError:
        logger.error('VersionOne plugin misconfigured, check your settings')
        raise
    except Exception:
        logger.exception('VersionOne failed to get client')
        raise

    return v1


def _get_review(item):
    for field in settings.VERSIONONE_CR_FIELDS:
        try:
            val = getattr(item, field)
            return '' if val is None else val, field
        except AttributeError:
            pass
    # No candidate fields matched, number might not have been a valid type
    raise NotFound


@deferred_to_channel
def review_command(v1, client, channel, nick, number, *args):
    """(review | cr) <issue> [!]<text>
       With no text, show review link
       With '!' before text replace link, otherwise append
    """

    w = get_workitem(v1, number)

    try:
        link, field = _get_review(w)
    except NotFound:
        # No candidate fields matched, number might not have been valid
        return 'I\'m sorry {0}, item "{1}" doesn\'t support reviews'.format(nick, number)

    if args is ():
        return '{0} Reviews: {1}'.format(number, link or '(None)')

    # else append CR
    change = False
    new_link = ' '.join(args)
    if new_link[0] == '!':
        new_link = new_link[1:]
        change = True
    elif new_link not in link:
        new_link = ' '.join([link, new_link])
        change = True

    if change:
        logger.debug('On {0} change {1} to "{2}"'.format(number, field, new_link))
        return commit_changes(v1, (w, field, new_link),)

    return 'Already got that one {0}'.format(nick)


@deferred_to_channel
def team_command(v1, client, channel, nick, *args):
    try:
        subcmd = args[0]
        args = args[1:]
    except IndexError:
        subcmd = 'list'

    # Find the named channel or create new
    q = {'name': channel}
    channel_settings = db.v1_channel_settings.find_one(q) or q
    teams = channel_settings.get('teams', {})
    # NB: White space is lost in command parsing, hope for the best
    name = ' '.join(args)

    if subcmd == 'list':
        return '\n'.join([
            '{0} {1}'.format(t, u) for t, u in teams.iteritems()
        ]) if teams else 'No teams found for {0}'.format(channel)
    elif subcmd == 'add':
        try:
            team = v1.Team.where(Name=name).first()
        except IndexError:
            return 'I\'m sorry {0}, team name "{1}" not found'.format(nick, name)
        # Manually building a url is lame, but the url property on TeamRooms doesn't work
        teams[name] = ', '.join([
            '{0}/TeamRoom.mvc/Show/{1}'.format(settings.VERSIONONE_URL, r.intid) for r in team.Rooms
        ]) or team.url
    elif subcmd == 'remove':
        try:
            del teams[name]
        except KeyError:
            return 'I\'m sorry {0}, team name "{1}" not found for {2}'.format(nick, name, channel)
    else:
        return 'No {0}, you can\'t {1}!'.format(nick, subcmd)
    # If we didn't return by now, save teams back to DB, and ack the user
    channel_settings['teams'] = teams
    db.v1_channel_settings.save(channel_settings)
    return random_ack()


@deferred_to_channel
def take_command(v1, client, channel, nick, number):
    w = get_workitem(v1, number, 'Owners')
    user = get_user(v1, nick)

    if user in w.Owners:
        return 'Dude {0}, you already own it!'.format(nick)
    # Writing to Owners can only add values
    return commit_changes(v1, (w, 'Owners', [user]))


def _list_or_add_things(v1, class_name, number, action=None, *args):
    Klass = getattr(v1, class_name)
    workitem = get_workitem(v1, number)
    if action is None:
        things = list(_get_things(Klass, workitem))
        things.sort(key=lambda t: t.Status.Order)

        return '\n'.join([
            '[{0}] {1} {2}'.format(t.Status.Name, t.Name, t.url)
            for t in things
        ]) if things else 'Didn\'t find any {0}s for {1}'.format(class_name, number)

    if action != u'add':
        raise QuitNow('I can\'t just "{0}" that, {{nick}}'.format(action))

    name = ' '.join(args)

    if not name:
        raise QuitNow('I\'m going to need a title for that, {nick}')

    t = create_object(
        Klass,
        Name=name,
        Parent=workitem.idref,
    )

    raise QuitNow('I created {0} {1} for you, {{nick}}'.format(t.Name, t.url))


@deferred_to_channel
def tasks_command(v1, client, channel, nick, number, action=None, *args):
    return _list_or_add_things(v1, 'Task', number, action, *args)


@deferred_to_channel
def tests_command(v1, client, channel, nick, number, action=None, *args):
    return _list_or_add_things(v1, 'Test', number, action, *args)


@deferred_to_channel
def user_command(v1, client, channel, nick, *args):
    # Recombine space'd args for full name lookup
    lookup = ' '.join(args) or nick
    user = get_user(v1, lookup)
    return '{0} [{1}] ({2})'.format(user.Name, user.Nickname, user.url)


def find_versionone_numbers(message):
    """
    Finds all versionone ticket numbers in a message. This will ignore any that already
    appear in a URL
    """

    pat = r'\b(({0})-\d+)\b'.format('|'.join(VERSIONONE_PATTERNS))
    tickets = []
    for ticket in re.findall(pat, message, re.IGNORECASE):
        tickets.append(ticket[0])

    return tickets


def versionone_command(v1, client, channel, nick, message, cmd, args):
    """
    Command handler for the versionone plugin
    """
    try:
        subcmd = args.pop(0)
    except IndexError:
        return [
            'Usage for versionone (alias v1)',
            '!v1 alias [lookup | set | remove] - Lookup an alias, or set/remove your own',
            '!v1 oauth [<code> | forget] - Configure or remove your oauth tokens',
            '!v1 review <issue> [!]<text> - Lookup, append, or set codereview field (alias: cr)',
            '!v1 take <ticket-id> - Add yourself to the ticket\'s Owners',
            '!v1 tasks <ticket-id> (add <title>) - List tasks for ticket, or add one',
            '!v1 team[s] [add | remove | list] <teamname> -- add, remove, list team(s) for the channel',
            '!v1 tests <ticket-id> (add <title>) - List tests for ticket, or add one',
            '!v1 token [<code> | forget] - Configure or remove your v1 API token',
            '!v1 user <nick> - Lookup V1 user for an ircnick',
        ]
    logger.debug('Calling VersionOne subcommand {0} with args {1}'.format(subcmd, args))

    try:
        return COMMAND_MAP[subcmd](v1, client, channel, nick, *args)
    except KeyError:
        return u'Umm... {0}, Never heard of it?'.format(subcmd)
    except TypeError:
        # can only be reached by a command that doesn't use the deffered_to_* wrapper
        return u'Umm... {0}, you might want to check the docs for {1}'.format(nick, subcmd)


@deferred_to_channel
def versionone_full_descriptions(v1, client, channel, nick, message, matches):
    """
    Meant to be run asynchronously because it uses the network
    """
    specials = defaultdict(list)

    for m in matches:
        # Build lists of special lookup types
        kind = m.split('-')[0]
        if kind in SPECIAL_PATTERNS:
            specials[SPECIAL_PATTERNS[kind]].append(m)
        else:
            # Or default to Workitem
            specials['Workitem'].append(m)

    descriptions = []
    for kind, vals in specials.items():
        descriptions += [
            u'[{number}] {name} ({url})'.format(**{
                'name': s.Name,
                'number': s.Number,
                'url': s.url,
            })
            # Use the right Endpoint
            for s in getattr(v1, kind).filter(
                # OR join on each number
                '|'.join(["Number='{0}'".format(n) for n in vals])
            ).select('Name', 'Number')
        ]

    return '\n'.join(descriptions)


@match(find_versionone_numbers)
@command('versionone', aliases=['v1'], help='Interact with VersionOne tickets.'
         'Usage: "!v1" for help')
def versionone(client, channel, nick, message, *args):
    """
    A plugin for showing URLs to VERSIONONE ticket numbers. This is both a set of commands to
    interact with tickets, and a match to automatically show them.

    Issue numbers are automatically detected.
    """

    try:
        v1 = get_v1(nick)
    except QuitNow:
        # With OAUTH, get_creds can raise QuitNow
        logger.warning('No v1 connection for {0}'.format(nick))
        v1 = None

    if len(args) == 2:
        # args = [cmd, args]
        fn = versionone_command
    else:
        # args = [matches]
        fn = versionone_full_descriptions
    res = fn(v1, client, channel, nick, message, *args)

    if isinstance(res, Deferred):
        raise ResponseNotReady
    return res


COMMAND_MAP = {
    'alias': alias_command,
    'cr': review_command,
    'oauth': oauth_command,
    'review': review_command,
    'take': take_command,
    'tasks': tasks_command,
    'team': team_command,
    'teams': team_command,
    'tests': tests_command,
    'token': token_command,
    'user': user_command,
}



from mock import patch

from .util import V1TestCase, writeable_settings_stub, settings_stub


class ReviewMixin(object):
    def setUp(self):
        super(ReviewMixin, self).setUp()
        self.cr_field = settings_stub.VERSIONONE_CR_FIELDS[0]
        self.link = 'http://example.com/code_review'
        self.new_link = 'http://another.link.com/'
        setattr(self.get_workitem(), self.cr_field, self.link)


class TestReviewCommand(ReviewMixin, V1TestCase):
    get_workitem = patch('helga_versionone.get_workitem')

    def test_problem(self):
        # Field names are wrong in settings, or looking for an object that
        # doesn't Actually support reviews
        w = self.get_workitem()
        for field in settings_stub.VERSIONONE_CR_FIELDS:
            delattr(w, field)

        return self._test_command(
            'cr whatever',
            'I\'m sorry {0}, item "whatever" doesn\'t support reviews'.format(self.nick),
        )

    def test_results(self):
        return self._test_command(
            'cr whatever',
            'whatever Reviews: {0}'.format(self.link),
        )

    def test_set_append_fail(self):
        return self._test_command(
            'review whatever {0}'.format(self.new_link),
            'I would, but I\'m not allowed to write :('
        )


class TestReviewCommandWithWrite(ReviewMixin, V1TestCase):
    settings = patch('helga_versionone.settings', writeable_settings_stub)
    get_workitem = patch('helga_versionone.get_workitem')

    def test_set_append(self):
        w = self.get_workitem()
        d = self._test_command(
            'cr whatever {0}'.format(self.new_link),
        )

        def check(res):
            # Check v1 attr, and commit called
            self.assertEquals(
                # self.v1.[w].[cr_field]
                getattr(w, self.cr_field),
                ' '.join([self.link, self.new_link]),
            )
            self.v1.commit.assert_called_once_with()
            self.assertAck()

        d.addCallback(check)
        return d

    def test_set_replace(self):
        w = self.get_workitem()
        d = self._test_command(
            'cr whatever !{0}'.format(self.new_link),
        )

        def check(res):
            # Check v1 attr, and commit called
            self.assertEquals(
                # self.v1.[w].[cr_field]
                getattr(w, self.cr_field),
                self.new_link,
            )
            self.v1.commit.assert_called_once_with()
            self.assertAck()

        d.addCallback(check)
        return d

    def test_set_ignored(self):
        return self._test_command(
            'cr whatever {0}'.format(self.link),
            'Already got that one {0}'.format(self.nick),
        )



from mock import patch

from .util import V1TestCase, writeable_settings_stub


class TestTakeCommand(V1TestCase):
    get_workitem = patch('helga_versionone.get_workitem')
    get_user = patch('helga_versionone.get_user')

    def test_in_owners(self):
        self.get_workitem().Owners = [self.nick, 'fhqwhgads']
        self.get_user.return_value = self.nick

        return self._test_command(
            'take whatever',
            'Dude {0}, you already own it!'.format(self.nick),
        )

    def test_not_in_owners_write_fail(self):
        self.get_workitem().Owners = ['fhqwhgads']
        self.get_user.return_value = self.nick

        return self._test_command(
            'take whatever',
            'I would, but I\'m not allowed to write :('
        )


class TestTakeCommandWithWrite(V1TestCase):
    settings = patch('helga_versionone.settings', writeable_settings_stub)
    get_workitem = patch('helga_versionone.get_workitem')
    get_user = patch('helga_versionone.get_user')

    def test_not_in_owners_write_ok(self):
        w = self.get_workitem()
        w.Owners = ['fhqwhgads']
        self.get_user.return_value = self.nick

        d = self._test_command(
            'take whatever',
        )

        def check(res):
            # Check data and commit called
            self.assertEquals(
                w.Owners,
                [self.nick],
            )
            self.v1.commit.assert_called_once_with()
            self.assertAck()

        d.addCallback(check)
        return d



from mock import call, patch
from pretend import stub

from oauth2client.client import FlowExchangeError

from .util import V1TestCase


class TestOauthCommandDisabled(V1TestCase):
    def test_disabled(self):
        self.USE_OAUTH = False

        return self._test_command(
            'oauth',
            'Oauth is not enabled',
            to_nick=True,
        )


class TestOauthCommand(V1TestCase):
    USE_OAUTH = patch('helga_versionone.USE_OAUTH', True)
    OAuth2WebServerFlow = patch('helga_versionone.OAuth2WebServerFlow')

    def test_get_code(self):
        url = 'http://example.com'
        self.OAuth2WebServerFlow().step1_get_authorize_url.return_value = url

        return self._test_command(
            'oauth',
            'Visit {0} then do "!v1 oauth <code>" with the generated code'.format(url),
            to_nick=True,
        )

    def test_reply_server_fail(self):
        msg = 'Whut!'
        self.OAuth2WebServerFlow().step2_exchange.side_effect = FlowExchangeError(msg)

        return self._test_command(
            'oauth somebiglongoauthreturncode',
            'Sorry {0} "{1}" happened. Try "!v1 oauth" again from the start'.format(
                self.nick, msg),
            to_nick=True,
        )

    def test_reply_ok(self):
        auth_info = {}
        auth_info['access_token'] = 'this is my token'
        auth_info['refresh_token'] = 'a refresh token'
        auth_info['token_expiry'] = 'sometime tomorrow'

        # No existing user
        self.db.v1_oauth.find_one.return_value = {}

        # Usable creds from Oauth
        self.OAuth2WebServerFlow().step2_exchange.return_value = stub(**auth_info)

        d = self._test_command(
            'oauth somebiglongoauthreturncode',
        )

        def check(res):
            # assertDictContainsSubset would be super cool, but it's a 2.7 feature
            # and we know what will be here
            auth_info.update({'irc_nick': self.nick})
            self.assertEqual(self.db.v1_oauth.save.call_args, call(auth_info))
            self.assertAck()

        d.addCallback(check)
        return d



from mock import patch, MagicMock
from unittest import TestCase
from httplib2 import HttpLib2ErrorWithResponse

from helga_versionone.v1_wrapper import property_required, HelgaOauthV1Server


class TestPropertyRequired(TestCase):
    def test_good_object_works(self):

        class A(object):
            foo = True

            @property_required('foo')
            def bar(self):
                return self.foo

        a = A()
        self.assertEqual(a.bar(), True)

    def test_lazy_object_works(self):

        class A(object):
            @property_required('foo')
            def bar(self):
                return self.foo

        a = A()
        a.foo = "I'm foo"
        self.assertEqual(a.bar(), "I'm foo")

    def test_broken_object_fails(self):

        class A(object):
            @property_required('foo')
            def bar(self):
                return "I don't really care about foo, but it is required"

        a = A()
        self.assertRaises(ValueError, a.bar)


class TestHelgaOauthV1Server(TestCase):
    def test_positional_args(self):
        s = HelgaOauthV1Server('example.com', 'instance_name')
        self.assertEquals(s.scheme, 'http')
        self.assertEquals(s.instance_url, 'http://example.com/instance_name/')

    def test_instance_url(self):
        s = HelgaOauthV1Server(instance_url='http://example.com/instance_name')
        self.assertEquals(s.scheme, 'http')
        self.assertEquals(s.address, 'example.com')
        self.assertEquals(s.instance, 'instance_name')

    def test_with_creds(self):
        creds = MagicMock()
        s = HelgaOauthV1Server(instance_url='http://example.com/instance_name', credentials=creds)
        creds.authorize.call_args.assert_called_once_with(s.httpclient)


class TestHelgaOauthV1ServerWithInstance(TestCase):
    def setUp(self):
        self.url = 'http://example.com/instance_name'
        self.server = HelgaOauthV1Server(instance_url=self.url)
        self.server.httpclient = MagicMock()
        self.server.httpclient.request.return_value = ('response', 'content')

    def test_fetch_get(self):
        e, c = self.server.fetch('/test')
        self.server.httpclient.request.call_args.assert_called_once_with(
            '{0}/test'.format(self.url), method='GET')
        self.assertEquals(e, None)
        self.assertEquals(c, 'content')

    def test_fetch_post(self):
        e, c = self.server.fetch('/test', postdata={'q': 'query'})
        self.server.httpclient.request.call_args.assert_called_once_with(
            '{0}/test'.format(self.url), method='POST', body='q=query')
        self.assertEquals(e, None)
        self.assertEquals(c, 'content')

    def test_fetch_get_fails(self):
        response = MagicMock()
        error = HttpLib2ErrorWithResponse('desc', response, 'error content')
        self.server.httpclient.request.side_effect = error
        e, c = self.server.fetch('/test')
        self.assertEqual(e, error)
        self.assertEqual(c, 'error content')
        self.assertEqual(e.response, response)



import logging

from copy import copy
from functools import wraps
from mock import MagicMock, patch
from mock.mock import _patch as Patch
from pretend import stub
from twisted.internet.defer import Deferred
from twisted.trial import unittest

import helga_versionone
from helga.plugins import ACKS


logger = logging.getLogger(__name__)
settings_stub = stub(
    VERSIONONE_URL='https://www.example.com/EnvKey',
    VERSIONONE_AUTH=('username', 'password'),
    VERSIONONE_CR_FIELDS=('field_one', 'field_two'),
    VERSIONONE_READONLY=True,
    VERSIONONE_OAUTH_ENABLED=False,
    VERSIONONE_OAUTH_CLIENT_ID='client_id',
    VERSIONONE_OAUTH_CLIENT_SECRET='client_secret',
)

writeable_settings_stub = copy(settings_stub)
writeable_settings_stub.VERSIONONE_READONLY = False


class deferred_patch(object):
    """Patches target until the deferred returns, appends patched object to args"""
    def __init__(self, target):
        self.patcher = patch(target)

    def __call__(self, fn):
        @wraps(fn)
        def inner(herself, *args, **kwargs):
            p = self.patcher.start()
            args = list(args)
            args.append(p)
            # patcher.stop() is always None, so return res, don't eat it (allows failures to propagate)
            return fn(herself, *args, **kwargs).addBoth(lambda res: self.patcher.stop() or res)

        return inner


def stop_patches(fn):
    @wraps(fn)
    def inner(self, *args, **kwargs):
        res = fn(self, *args, **kwargs)
        if isinstance(res, Deferred):
            # clearPatches() is always None, so return x, don't eat it (allows failures to propagate)
            return res.addBoth(lambda x: self.clearPatches() or x)
        # Not deferred, stop patches and return
        self.clearPatches()
        return res
    return inner


class PatchedTestCaseMeta(type):
    def __new__(meta, name, bases, dct):
        """Decorate all test methods in dct"""

        for k, v in dct.iteritems():
            # Named test* and is a method
            if k.startswith('test') and callable(v):
                logger.debug('decorate {0} -> {1} for {2}'.format(k, v, meta))
                # Then decorate with the stopper
                dct[k] = stop_patches(v)
                dct[k].__patched = True

        return super(PatchedTestCaseMeta, meta).__new__(meta, name, bases, dct)

    def __init__(cls, name, bases, dct):
        """Collect patches"""
        if hasattr(cls, '_mock_patches'):
            # This must be from a base class,
            # copy it so it doesn't get mutated in the base
            cls._mock_patches = copy(cls._mock_patches)
        else:
            cls._mock_patches = set()

        for k, v in dct.iteritems():
            if isinstance(v, Patch):
                # If it's a patch add it to patches
                logger.debug('Adding {0} patch to {1}'.format(k, cls))
                cls._mock_patches.add(k)

        return super(PatchedTestCaseMeta, cls).__init__(name, bases, dct)


class PatchedTestCase(unittest.TestCase):
    """class properties which are mock._patch instances (returned by mock.patch) will become
       instance variables which are the mock object.

       patches are started in setUp

       patches are stopped when testMethod completes, either normally or through a deferred
    """

    __metaclass__ = PatchedTestCaseMeta

    def setUp(self):
        # Before each test, start the patches
        # metaclass has decorated the test methods to call clearPatches
        for attr in self._mock_patches:
            patcher = getattr(self, attr)
            setattr(self, '_{0}_patcher'.format(attr), patcher)
            setattr(self, attr, patcher.start())

        super(PatchedTestCase, self).setUp()

    def clearPatches(self):
        for p in self._mock_patches:
            getattr(self, '_{0}_patcher'.format(p)).stop()


class V1TestCase(PatchedTestCase):
    """Base class for all helga_versionone tests
       helga_versionone.get_v1 is always patched to return self.v1 - a MagicMock()
       helga_versionone.settings is always patched to settings_stub
       helga_versionone.db is always patched
    """

    nick = 'me'
    channel = '#bots'

    db = patch('helga_versionone.db')
    settings = patch('helga_versionone.settings', settings_stub)
    get_v1 = patch('helga_versionone.get_v1')

    def setUp(self):
        # Starts the patches
        super(V1TestCase, self).setUp()

        self.v1 = MagicMock()
        # Depends on "get_v1" being in patches above
        self.get_v1.return_value = self.v1

        # client is mocked, but doesn't have to be patched.
        self.client = MagicMock()

    def assertAck(self):
        self.assertIn(self.client.msg.call_args[0][1], ACKS)

    def _test_command(self, cmd, expected=None, to_nick=False):
        """Test a version command: !v1 <cmd>"""

        d = helga_versionone.versionone_command(
            self.v1,
            self.client,
            self.channel,
            self.nick,
            'unused message',
            'unused cmd',
            cmd.split(),
        )

        if expected is None:
            return d

        if isinstance(d, Deferred):
            target = self.nick if to_nick else self.channel

            def check(actual):
                self.client.msg.assert_called_once_with(target, expected)

            d.addCallback(check)
            return d
        else:
            self.assertEqual(d, expected)



from mock import patch
from pretend import stub

from .util import V1TestCase, writeable_settings_stub

# Tests for both tests and tasks as they are the same code path


def make_thing(name, num, status_name, status_order):
    return stub(
        Name='{0}-{1}'.format(name, num),
        url='http://example.com/{0}'.format(num),
        Status=stub(
            Name=status_name,
            Order=status_order,
        )
    )


class TestThingCommand(V1TestCase):
    get_workitem = patch('helga_versionone.get_workitem')

    def setUp(self):
        super(TestThingCommand, self).setUp()
        self.results = [
            make_thing('thing', 1, 'Done', 99),
            make_thing('thing', 2, 'None', 0),
            make_thing('thing', 3, 'In Progress', 50),
        ]
        self.results_ordered = [self.results[i] for i in [1, 2, 0]]
        self.v1.Task.where().select.return_value = self.results

    def test_list_tasks(self):
        return self._test_command(
            'tasks whatever',
            '\n'.join([
                '[{0}] {1} {2}'.format(t.Status.Name, t.Name, t.url)
                for t in self.results_ordered
            ]),
        )

    def test_list_tests_none(self):
        return self._test_command(
            'tests whatever',
            'Didn\'t find any Tests for whatever',
        )

    def test_bad_action(self):
        return self._test_command(
            'tests whatever fhqwhgads',
            'I can\'t just "fhqwhgads" that, {0}'.format(self.nick),
        )

    def test_add_failed_for_title(self):
        return self._test_command(
            'tests whatever add',
            'I\'m going to need a title for that, {0}'.format(self.nick),
        )

    def test_add_failed_no_write(self):
        return self._test_command(
            'tasks whatever add Do a little dance',
            'I\'m sorry {0}, write access is disabled'.format(self.nick),
        )


class TestThingCommandWithWrite(V1TestCase):
    settings = patch('helga_versionone.settings', writeable_settings_stub)
    get_workitem = patch('helga_versionone.get_workitem')

    def test_tests_add_ok(self):
        self.get_workitem().idref = 3
        name = 'Do a little dance'
        url = 'http://example.com'

        self.v1.Test.create.return_value = stub(
            Name=name,
            url=url,
        )

        d = self._test_command(
            'tests whatever add {0}'.format(name),
            'I created {0} {1} for you, {2}'.format(name, url, self.nick)
        )

        def check(res):
            # Check data and commit called
            self.v1.Test.create.assert_called_once_with(
                Name=name,
                Parent=3,
            )

        d.addCallback(check)
        return d



from mock import MagicMock, patch
from pretend import stub

import helga_versionone

from .util import V1TestCase


class TestCommands(V1TestCase):
    def test_no_patching(self):
        # settings are patched
        from helga_versionone import settings
        assert settings.VERSIONONE_URL == 'https://www.example.com/EnvKey'

    def test_patching(self):
        # settings can be overridden
        self.settings.VERSIONONE_URL = 'tada'
        assert helga_versionone.settings.VERSIONONE_URL == 'tada'

    def test_bad_command(self):
        return self._test_command('notreal', u'Umm... notreal, Never heard of it?')

    def test_get_workitem_found(self):
        self.v1.Workitem.where().select().first.return_value = 'foo'
        f = helga_versionone.get_workitem(self.v1, 'B-00010')
        self.assertEquals(f, 'foo')

    def test_get_workitem_not_found(self):
        self.v1.Workitem.where.side_effect = IndexError
        self.assertRaises(helga_versionone.QuitNow, helga_versionone.get_workitem, self.v1, 'B-00010')

    def test_get_user_found(self):
        self.v1.Member.filter().select().first.return_value = 'foo'
        u = helga_versionone.get_user(self.v1, self.nick)
        self.assertEquals(u, 'foo')

    def test_get_user_not_found(self):
        self.v1.Member.filter.side_effect = IndexError
        self.assertRaises(helga_versionone.QuitNow, helga_versionone.get_user, self.v1, self.nick)

    def test_get_user_with_no_alias(self):
        self.db.v1_user_map.find_one.side_effect = KeyError
        self.v1.Member.filter().select().first.return_value = 'foo'
        u = helga_versionone.get_user(self.v1, self.nick)
        self.assertEquals(u, 'foo')

    @patch('helga_versionone.OAuth2Credentials')
    @patch('helga_versionone.USE_OAUTH')
    def _test_get_creds(self, name, mock_USE_OAUTH, mock_OAuth2Credentials, oauth_works=False, token_works=True):
        # First call should fail, so nick re-writing happens
        # Second needs keys for the call to OAuth2Credentials
        auth_info = {}
        mock_USE_OAUTH = False

        if oauth_works:
            mock_USE_OAUTH = True
            auth_info['access_token'] = 'asdf'
            auth_info['refresh_token'] = 'asdf'
            auth_info['token_expiry'] = 'asdf'

        if token_works:
            auth_info['api_token'] = 'mahtoken'

        self.db.v1_oauth.find_one.side_effect = [None, auth_info]
        if oauth_works:
            mock_OAuth2Credentials.return_value = 'foo'
        else:
            mock_OAuth2Credentials.side_effect = KeyError

        if oauth_works and not token_works:
            c = helga_versionone.get_creds(name)
            self.assertEquals(c, 'foo')
        elif token_works:
            c = helga_versionone.get_creds(name)
            self.assertEquals(c, 'mahtoken')
        else:
            self.assertRaises(helga_versionone.QuitNow, helga_versionone.get_creds, name)

    def test_get_creds_plain(self):
        self._test_get_creds('somename')

    def test_get_creds_underscore(self):
        # No pipe - split underscore
        self._test_get_creds('somename_away')
        self.db.v1_oauth.find_one.called_with({'irc_nick': 'somename'})

    def test_get_creds_pipe(self):
        # With pipe - split pipe only
        self._test_get_creds('some_name|away')
        self.db.v1_oauth.find_one.called_with({'irc_nick': 'some_name'})

    def test_get_creds_oauth_works(self):
        self._test_get_creds('fhqwhgads', oauth_works=True, token_works=False)

    def test_get_creds_fail_all(self):
        self._test_get_creds('fhqwhgads', oauth_works=False, token_works=False)

    def test_find_all(self):
        numbers = helga_versionone.find_versionone_numbers('Tell me about B-0010')
        self.assertEquals(numbers, ['B-0010'])

    def test_usage(self):
        res = self._test_command('')
        self.assertIn('Usage for versionone', res[0])

    def test_bad_arg_count(self):
        return self._test_command(
            'take',
            u'Umm... {0}, you might want to check the docs for that'.format(self.nick),
        )

    @patch('helga_versionone.versionone_full_descriptions')
    def test_plugin_match(self, fn):
        msg = 'Tell me about B-0010',
        matches = ['B-0010']
        helga_versionone.versionone(
            self.client,
            self.channel,
            self.nick,
            msg,
            matches,
        )

        fn.assert_called_once_with(
            self.v1,
            self.client,
            self.channel,
            self.nick,
            msg,
            matches,
        )

    @patch('helga_versionone.versionone_command')
    def test_plugin_subcommand(self, fn):
        # QuitNow doesn't stop processing of command
        self.get_v1.side_effect = helga_versionone.QuitNow()

        msg = '!v1 teams'
        cmd = '!v1'
        args = ['teams']

        helga_versionone.versionone(
            self.client,
            self.channel,
            self.nick,
            msg,
            cmd,
            args,
        )

        fn.assert_called_once_with(
            None,
            self.client,
            self.channel,
            self.nick,
            msg,
            cmd,
            args,
        )

    def test_no_v1_for_command(self):
        self.v1 = None
        # For coverage, hit bad_args with v1 == None
        return self._test_command(
            'tasks B-0010 fhqwhgads',
            u'{0}, you might want to try "!v1 oauth" or "!v1 token"'.format(self.nick),
        )

    def test_versionone_full_descriptions(self):
        w = stub(
            Name='Issue name',
            Number='B-0010',
            url='http://example.com',
        )
        self.v1.Workitem.filter().select.return_value = [w]

        d = helga_versionone.versionone_full_descriptions(
            self.v1,
            self.client,
            self.channel,
            self.nick,
            'Something about B-0010',
            ['B-0010'],
        )

        def check(res):
            self.client.msg.assert_called_once_with(
                self.channel,
                '[{number}] {name} ({url})'.format(**{
                    'name': w.Name,
                    'number': w.Number,
                    'url': w.url,
                })
            )

        d.addCallback(check)
        return d

    def test_versionone_full_descriptions_special(self):
        w = stub(
            Name='Issue name',
            Number='I-0010',
            url='http://example.com',
        )
        self.v1.Issue.filter().select.return_value = [w]

        d = helga_versionone.versionone_full_descriptions(
            self.v1,
            self.client,
            self.channel,
            self.nick,
            'Something about I-0010',
            ['I-0010'],
        )

        def check(res):
            self.client.msg.assert_called_once_with(
                self.channel,
                '[{number}] {name} ({url})'.format(**{
                    'name': w.Name,
                    'number': w.Number,
                    'url': w.url,
                })
            )

        d.addCallback(check)
        return d


class TestUserCommand(V1TestCase):
    get_user = patch('helga_versionone.get_user', return_value=stub(
        Name='fhqwhgads',
        Nickname='joe',
        url='http://example.com',
    ))

    def test_user_default_command(self):
        d = self._test_command(
            'user',
            'fhqwhgads [joe] (http://example.com)',
        )

        def check(res):
            self.get_user.assert_called_once_with(self.v1, self.nick)

        d.addCallback(check)
        return d

    def test_user_explicit_command(self):
        d = self._test_command(
            'user fhqwhgads',
            'fhqwhgads [joe] (http://example.com)',
        )

        def check(res):
            self.get_user.assert_called_once_with(self.v1, 'fhqwhgads')

        d.addCallback(check)
        return d



from mock import patch

from .util import V1TestCase


class TestTokenCommand(V1TestCase):
    get_workitem = patch('helga_versionone.get_workitem')
    get_user = patch('helga_versionone.get_user')

    def test_step1(self):
        auth_info = {}
        self.db.v1_oauth.find_one.return_value = auth_info

        return self._test_command(
            'token',
            'In V1 go to your Applications and generate a Personal Access Token'
            'then do "!v1 token <code>" with the generated code',
            to_nick=True
        )
    def test_no_token(self):
        auth_info = {}
        self.db.v1_oauth.find_one.return_value = auth_info

        return self._test_command(
            'token forget',
            'Token was already gone',
            to_nick=True
        )

    def test_forget_works(self):
        auth_info = {'api_token': 'delete me'}
        self.db.v1_oauth.find_one.return_value = auth_info

        d = self._test_command(
            'token forget',
            to_nick=True
        )

        def check(res):
            self.assertEqual(auth_info, {})
            self.assertAck()

        d.addCallback(check)
        return d

    def test_set_token(self):
        # If the return were False, a new dict gets created
        auth_info = {'notfalse': True}
        self.db.v1_oauth.find_one.return_value = auth_info

        d = self._test_command(
            'token newvalue',
            to_nick=True
        )

        def check(res):
            self.assertEqual(auth_info, {'api_token': 'newvalue', 'notfalse': True})
            self.assertAck()

        d.addCallback(check)
        return d



from mock import MagicMock
from pretend import stub

from .util import V1TestCase


class TestTeamCommand(V1TestCase):
    def test_no_teams(self):
        self.db.v1_channel_settings.find_one.return_value = None
        return self._test_command('teams', u'No teams found for {0}'.format(self.channel))

    def test_teams(self):
        self.db.v1_channel_settings.find_one.return_value = {'teams': {'teamName': 'link'}}
        return self._test_command('teams', u'teamName link')

    def test_team_add_fail(self):
        self.v1.Team.where.side_effect = IndexError
        return self._test_command(
            'teams add not there',
            u'I\'m sorry {0}, team name "not there" not found'.format(self.nick),
        )

    def test_team_add_ok_url(self):
        self.db.v1_channel_settings.find_one.return_value = None
        team = MagicMock()
        team.url = 'http://example.com/'
        team.Rooms = []
        self.v1.Team.where().first.return_value = team
        d = self._test_command(
            'teams add team name',
        )

        def check(res):
            # call_args[0] ==> *args
            self.assertEqual(self.db.v1_channel_settings.save.call_args[0][0]['teams']['team name'], team.url)
            self.assertAck()

        d.addCallback(check)
        return d

    def test_team_add_ok_rooms(self):
        self.db.v1_channel_settings.find_one.return_value = None
        team = MagicMock()
        team.Rooms = [stub(intid=3)]
        self.v1.Team.where().first.return_value = team
        d = self._test_command(
            'teams add team name',
        )

        def check(res):
            self.assertEqual(
                self.db.v1_channel_settings.save.call_args[0][0]['teams']['team name'],
                '{0}/TeamRoom.mvc/Show/3'.format(self.settings.VERSIONONE_URL),
            )
            self.assertAck()

        d.addCallback(check)
        return d

    def test_team_remove_fail(self):
        self.db.v1_channel_settings.find_one.return_value = None
        return self._test_command(
            'team remove team name',
            'I\'m sorry {0}, team name "team name" not found for {1}'.format(self.nick, self.channel),
        )

    def test_team_remove_ok(self):
        self.db.v1_channel_settings.find_one().get.return_value = {'this one': 'http://example.com'}
        d = self._test_command(
            'team remove this one',
        )

        d.addCallback(lambda _: self.assertAck())
        return d

    def test_team_no_command(self):
        return self._test_command(
            'team naugty',
            'No {0}, you can\'t naugty!'.format(self.nick),
        )



from setuptools import setup, find_packages


version = '0.1.9'

setup(name="helga-versionone",
      version=version,
      description=('VersionOne interface for helga chat bot'),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Topic :: Communications :: Chat :: Internet Relay Chat',
          'Framework :: Twisted',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords='helga versionone',
      author='Aaron McMillin',
      author_email='aaron@mcmillinclan.org',
      url='https://github.com/aarcro/helga-versionone',
      download_url='https://github.com/aarcro/helga-versionone/tarball/' + version,
      license='MIT',
      packages=find_packages(),
      py_modules=['helga_versionone'],
      install_requires=[
          'v1pysdk-unofficial',
          'oauth2client',
          'expiringdict',
      ],
      entry_points = dict(
          helga_plugins=[
              'versionone = helga_versionone:versionone'
          ],
      ),
)



from mock import call

from .util import V1TestCase


class TestAliasCommand(V1TestCase):
    def test_no_args_no_match(self):
        # v1_nick = db.v1_user_map.find_one(lookup)['v1_nick']
        self.db.v1_user_map.find_one.return_value = {}
        return self._test_command(
            'alias',
            '{0} is known as {1} in V1'.format(self.nick, self.nick),
        )

    def test_lookup_no_match(self):
        # v1_nick = db.v1_user_map.find_one(lookup)['v1_nick']
        self.db.v1_user_map.find_one.return_value = {}
        return self._test_command(
            'alias lookup',
            '{0} is known as {1} in V1'.format(self.nick, self.nick),
        )

    def test_no_args_match(self):
        # v1_nick = db.v1_user_map.find_one(lookup)['v1_nick']
        self.db.v1_user_map.find_one.return_value = {'v1_nick': 'nickname'}
        return self._test_command(
            'alias',
            '{0} is known as nickname in V1'.format(self.nick),
        )

    def test_lookup_match(self):
        # v1_nick = db.v1_user_map.find_one(lookup)['v1_nick']
        self.db.v1_user_map.find_one.return_value = {'v1_nick': 'nickname'}
        return self._test_command(
            'alias lookup',
            '{0} is known as nickname in V1'.format(self.nick),
        )

    def test_lookup_is_default(self):
        # v1_nick = db.v1_user_map.find_one(lookup)['v1_nick']
        self.db.v1_user_map.find_one.return_value = {}
        return self._test_command(
            'alias fhqwhgads',
            'fhqwhgads is known as fhqwhgads in V1',
        )

    def test_lookup_other__no_match(self):
        # v1_nick = db.v1_user_map.find_one(lookup)['v1_nick']
        self.db.v1_user_map.find_one.return_value = {}
        return self._test_command(
            'alias lookup fhqwhgads',
            'fhqwhgads is known as fhqwhgads in V1',
        )

    def test_lookup_other_match(self):
        # v1_nick = db.v1_user_map.find_one(lookup)['v1_nick']
        self.db.v1_user_map.find_one.return_value = {'v1_nick': 'nickname'}
        return self._test_command(
            'alias lookup fhqwhgads',
            'fhqwhgads is known as nickname in V1',
        )

    def test_set(self):
        # alias = db.v1_user_map.find_one(lookup) or lookup
        # db.v1_user_map.save(alias)
        self.db.v1_user_map.find_one.return_value = None

        d = self._test_command(
            'alias set fhqwhgads',
        )

        def check(res):
            self.assertEquals(
                self.db.v1_user_map.save.call_args,
                call({'irc_nick': self.nick, 'v1_nick': 'fhqwhgads'}),
            )
            self.assertAck()

        d.addCallback(check)
        return d

    def test_remove_implicit_ok(self):
        # alias = db.v1_user_map.find_one(lookup) or lookup
        self.db.v1_user_map.find_one.return_value = None

        d = self._test_command(
            'alias remove',
        )

        def check(res):
            self.assertEquals(
                self.db.v1_user_map.find_and_modify.call_args,
                call({'irc_nick': self.nick}, remove=True),
            )
            self.assertAck()

        d.addCallback(check)
        return d

    def test_remove_explicit_ok(self):
        # alias = db.v1_user_map.find_one(lookup) or lookup
        self.db.v1_user_map.find_one.return_value = None

        d = self._test_command(
            'alias remove {0}'.format(self.nick),
        )

        def check(res):
            self.assertEquals(
                self.db.v1_user_map.find_and_modify.call_args,
                call({'irc_nick': self.nick}, remove=True),
            )
            self.assertAck()

        d.addCallback(check)
        return d

    def test_remove_explicit_fail(self):
        return self._test_command(
            'alias remove fhqwhgads',
            'That\'s not nice {0}. You can\'t remove fhqwhgads'.format(self.nick)
        )

    def test_no_command(self):
        return self._test_command(
            'alias naughty fhqwhgads'
            'No {0}, you can\'t naughty!'.format(self.nick)
        )



"""Patch v1pysdk to use non-file credential stores"""

import logging
import httplib2
import urllib2

from urllib import urlencode
from urllib2 import HTTPBasicAuthHandler, HTTPCookieProcessor

from expiringdict import ExpiringDict
from functools import wraps
from urlparse import urlparse
from v1pysdk.client import V1Server
from v1pysdk.v1meta import V1Meta

try:
    from xml.etree import ElementTree
except ImportError:  # pragma: no cover
    from elementtree import ElementTree


logger = logging.getLogger(__name__)


class property_required(object):
    """Decorate a method to require certian properties be populated"""

    def __init__(self, *args):
        """Arguments to the decorator are proprties which must be non-None"""
        self.props = args

    def __call__(self, fn):
        """Called once durring decoration, returns the real func to be called"""

        @wraps(fn)
        def wrapped_fn(herself, *args, **kwargs):
            for prop in self.props:
                if getattr(herself, prop, None) is None:
                    msg = 'Required property {0} is missing from {1}'.format(prop, herself)
                    logger.error(msg)
                    raise ValueError(msg)
            return fn(herself, *args, **kwargs)

        return wrapped_fn


class HelgaV1Meta(V1Meta):  # pragma: no cover
    def __init__(self, *args, **kw):
        # Coppied from V1Meta, but use our own Server class
        self.server = HelgaOauthV1Server(*args, **kw)
        # ...And a cache that expires
        self.global_cache = ExpiringDict(max_len=100, max_age_seconds=10)
        self.dirtylist = []

    def asset_class(self, asset_type_name):
        Klass = super(HelgaV1Meta, self).asset_class(asset_type_name)
        # Always use the current V1Meta instance, not the closed in one
        Klass._v1_v1meta = self
        return Klass


class HelgaOauthV1Server(V1Server):
    "Accesses a V1 HTTP server as a client of the XML API protocol"
    API_PATH = "/rest-1.oauth.v1"

    def __init__(self, address="localhost", instance="VersionOne.Web", password='',
                 scheme="http", instance_url=None, credentials=None, use_password_as_token=False):
        # How hacky is this?
        self.logger = logging.getLogger(__name__ + '.v1_client')
        self.logger.setLevel(logging.INFO)
        # Do not make super call, base implementation requires username/password
        if instance_url:
            self.instance_url = instance_url
            parsed = urlparse(instance_url)
            self.address = parsed.netloc
            self.instance = parsed.path.strip('/')
            self.scheme = parsed.scheme
        else:
            self.address = address
            self.instance = instance.strip('/')
            self.scheme = scheme
            self.instance_url = self.build_url('')

        self.httpclient = None

        if credentials is not None:
            self.set_credentials(credentials)

        # Cheating to support tokens, since I know the internals
        if use_password_as_token:
            self.username = ''
            self.password = password
            self.use_password_as_token = use_password_as_token
            self._install_opener()
            # Become the parent class (for opener style get/post methods
            self.__class__ = V1Server

    def _install_opener(self):
        base_url = self.build_url('')
        password_manager = urllib2.HTTPPasswordMgrWithDefaultRealm()
        password_manager.add_password(None, base_url, self.username, self.password)
        # only use the first Auth_handler to avoid the busted NTLM handler
        self.opener = urllib2.build_opener(HTTPBasicAuthHandler(password_manager))
        if self.use_password_as_token:
            self.opener.addheaders.append(('Authorization', 'Bearer ' + self.password))
        self.opener.add_handler(HTTPCookieProcessor())
        # No attempt is made to clear the cache when the creds change.
        # We assume all users have the same read-access.

    def set_credentials(self, creds):
        # If there are memory leaks, they might come from here
        self.httpclient = httplib2.Http()
        creds.authorize(self.httpclient)

    @property_required('httpclient')
    def http_get(self, url):
        return self.httpclient.request(url, method='GET')  # pragma: no cover

    @property_required('httpclient')
    def http_post(self, url, data=''):
        return self.httpclient.request(url, method='POST', body=data)  # pragma: no cover

    def fetch(self, path, query='', postdata=None):
        "Perform an HTTP GET or POST depending on whether postdata is present"
        url = self.build_url(path, query=query)
        try:
            if postdata is not None:
                if isinstance(postdata, dict):
                    postdata = urlencode(postdata)
                response, body = self.http_post(url, postdata)
            else:
                response, body = self.http_get(url)
            return (None, body)
        except httplib2.HttpLib2ErrorWithResponse, e:
            if e.response.status == 401:
                raise  # pragma: no cover
            body = e.content
            return (e, body)

    def get_asset_xml(self, asset_type_name, oid):
        path = self.API_PATH + '/Data/{0}/{1}'.format(asset_type_name, oid)
        return self.get_xml(path)

    def get_query_xml(self, asset_type_name, where=None, sel=None):
        path = self.API_PATH + '/Data/{0}'.format(asset_type_name)
        query = {}
        if where is not None:
                query['Where'] = where
        if sel is not None:
                query['sel'] = sel
        return self.get_xml(path, query=query)

    def execute_operation(self, asset_type_name, oid, opname):
        path = self.API_PATH + '/Data/{0}/{1}'.format(asset_type_name, oid)
        query = {'op': opname}
        return self.get_xml(path, query=query, postdata={})

    def get_attr(self, asset_type_name, oid, attrname):
        path = self.API_PATH + '/Data/{0}/{1}/{2}'.format(asset_type_name, oid, attrname)
        return self.get_xml(path)

    def create_asset(self, asset_type_name, xmldata, context_oid=''):
        body = ElementTree.tostring(xmldata, encoding="utf-8")
        query = {}
        if context_oid:
            query = {'ctx': context_oid}
        path = self.API_PATH + '/Data/{0}'.format(asset_type_name)
        return self.get_xml(path, query=query, postdata=body)

    def update_asset(self, asset_type_name, oid, update_doc):
        newdata = ElementTree.tostring(update_doc, encoding='utf-8')
        path = self.API_PATH + '/Data/{0}/{1}'.format(asset_type_name, oid)
        return self.get_xml(path, postdata=newdata)
